/*
 *	Objective-C runtime type information
 *	provides compatibility between OSX/Darwin
 *	and GNUstep
 *
 *	written by Alexander V. Diemand
 *	and explicitly put into public domain: 
 *	you may redistribute, modify and alter this at your will.
 *
 */

#import <Foundation/Foundation.h>


#ifdef __APPLE__

#import <objc/objc-class.h>

#include "Darwin-GS_compatibility.h"


static int expfield[] = { 1,10,100,1000,10000,100000,1000000 };


char *objc_skip_typespec(const char *type)
{
    switch(*type)
    {
        case 'r': //            'r' == const
            return objc_skip_typespec (type+1);
            break;
        case _C_ID: //		'@'
        case _C_CLASS: //	'#'
        case _C_SEL: //		':'
        case _C_CHR: //		'c'
        case _C_UCHR: //	'C'
        case _C_SHT: //		's'
        case _C_USHT: //	'S'
        case _C_INT: //		'i'
        case _C_UINT: //	'I'
        case _C_LNG: //		'l'
        case _C_ULNG: //        'L'
        case _C_FLT: //		'f'
        case _C_DBL: //		'd'
            return type+1; break;
        case _C_BFLD: //        'b'
            [NSException raise:@"Primitive's Size" format:@"don't know about bitfield."];
            return type+1;
        case _C_VOID: //        'v'
        case _C_UNDEF: //	'?'
            return type+1; break;
        case _C_PTR: //		'^'
            return objc_skip_typespec(type+1); break;
        case _C_CHARPTR: //	'*'
            return type+1;
        case _C_ARY_B: //	'['
            {
                while (*type != '\0')
                {
                    if (*type == _C_ARY_E)
                    {
                        type++;
                        break;
                    }
                    type++;
                }
                return type;
            }
        case _C_ARY_E: //	']'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of the end of an array."];
            return type;
        case _C_UNION_B: //	'('
            {
                while (*type != '\0')
                {
                    if (*type == _C_UNION_E)
                    {
                        type++;
                        break;
                    }
                    type++;
                }
                return type;
            }
        case _C_UNION_E: //	')'
            [NSException raise:@"Primitive's Size" format:@"don't know about a union."];
            return type;
        case _C_STRUCT_B: //	'{'
            {
                while (*type != '\0')
                {
                    if (*type == _C_STRUCT_E)
                    {
                        type++;
                        break;
                    }
                    type++;
                }
                return type;
            }
        case _C_STRUCT_E: //	'}'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of the end of a structure."];
            return type;
        default:
	    printf("unknown type: %s\n",type);
            return type;
    }       
    return NULL;
}


int objc_sizeof_type (const char *type)
{
	//printf("objc_sizeof_type: %s\n",type);
    switch(*type)
    {
        case 'r': //            'r' == const
            return objc_sizeof_type (type+1);
            break;
        case _C_ID: //		'@'
            return sizeof(id);
        case _C_CLASS: //	'#'
            return sizeof(Class);
        case _C_SEL: //		':'
            return sizeof(SEL);
        case _C_CHR: //		'c'
            return sizeof(char);
        case _C_UCHR: //	'C'
            return sizeof(unsigned char);
        case _C_SHT: //		's'
            return sizeof(short);
        case _C_USHT: //	'S'
            return sizeof(unsigned short);
        case _C_INT: //		'i'
            return sizeof(int);
        case _C_UINT: //	'I'
            return sizeof(unsigned int);
        case _C_LNG: //		'l'
            return sizeof(long);
        case _C_ULNG: //        'L'
            return sizeof(unsigned long);
        case _C_FLT: //		'f'
            return sizeof(float);
        case _C_DBL: //		'd'
            return sizeof(double);
        case _C_BFLD: //        'b'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of a bitfield."];
            return sizeof(id);
        case _C_VOID: //        'v'
            return sizeof(void);
        case _C_UNDEF: //	'?'
            return sizeof(id);
        case _C_PTR: //		'^'
		{
			//int subtype = objc_sizeof_type(type+1);
			return sizeof(void*);
		}
        case _C_CHARPTR: //	'*'
            return sizeof(char*);
        case _C_ARY_B: //	'['
		{
			const char *stype = type+1;
			int counter;
			//int subtype;
			int cntlen;
			counter = 0;
			cntlen = 0;
			type++;
			while (*type >= '0' && *type <= '9')
			{
				cntlen++; type++;
			}
			//printf("cntlen=%d\n",cntlen);
			while (cntlen>0)
			{
				cntlen--;
				//printf("    at %d: %c\n",cntlen,stype[type-stype-cntlen-1]);
				counter += expfield[cntlen]*(stype[type-stype-cntlen-1]-48);
			}
			//printf("alen=%d\n",counter);
			return (counter * objc_sizeof_type(type));
		}
        case _C_ARY_E: //	']'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of the end of an array."];
            return sizeof(id);
        case _C_UNION_B: //	'('
            [NSException raise:@"Primitive's Size" format:@"don't know the size of a union."];
             return sizeof(id);
        case _C_UNION_E: //	')'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of a union."];
            return sizeof(id);
        case _C_STRUCT_B: //	'{'
            {
                int size = 0;
		/* first skip over name= */
                while (*type != '\0' && *type != _C_STRUCT_E)
		{
			//printf("    name=%s\n",type);
			if (*type == '=')
			{
				type++;
				break;
			}
			type++;
		}
		/* sum up all sizes of single types */
                while (*type != '\0' && *type != _C_STRUCT_E)
                {
			//printf("    type=%s\n",type);
			size += objc_sizeof_type(type);
			type++;
                }
                return size;
            }
        case _C_STRUCT_E: //	'}'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of the end of a structure."];
            return sizeof(id);
        default:
	    printf("unknown type: %s\n",type);
            return sizeof(id);
    }       
    return -1;
}


int objc_alignof_type (const char *type)
{
	//printf("objc_alignof_type: %s\n",type);
    switch(*type)
    {
        case 'r': //            'r' == const
            return objc_alignof_type (type+1);
            break;
        case _C_ID: //		'@'
            return __alignof__(id);
        case _C_CLASS: //	'#'
            return __alignof__(Class);
        case _C_SEL: //		':'
            return __alignof__(SEL);
        case _C_CHR: //		'c'
            return __alignof__(char);
        case _C_UCHR: //	'C'
            return __alignof__(unsigned char);
        case _C_SHT: //		's'
            return __alignof__(short);
        case _C_USHT: //	'S'
            return __alignof__(unsigned short);
        case _C_INT: //		'i'
            return __alignof__(int);
        case _C_UINT: //	'I'
            return __alignof__(unsigned int);
        case _C_LNG: //		'l'
            return __alignof__(long);
        case _C_ULNG: //        'L'
            return __alignof__(unsigned long);
        case _C_FLT: //		'f'
            return __alignof__(float);
        case _C_DBL: //		'd'
            return __alignof__(double);
        case _C_BFLD: //        'b'
            [NSException raise:@"Primitive's Size" format:@"don't know about bitfield."];
            return __alignof__(id);
        case _C_VOID: //        'v'
            return __alignof__(void);
        case _C_UNDEF: //	'?'
            return __alignof__(id);
        case _C_PTR: //		'^'
	    return __alignof__(void*);
        case _C_CHARPTR: //	'*'
            return __alignof__(char*);
        case _C_ARY_B: //	'['
	    return __alignof__(void*);
        case _C_ARY_E: //	']'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of the end of an array."];
            return __alignof__(id);
        case _C_UNION_B: //	'('
            [NSException raise:@"Primitive's Size" format:@"don't know about a union."];
             return __alignof__(id);
        case _C_UNION_E: //	')'
            [NSException raise:@"Primitive's Size" format:@"don't know about a union."];
            return __alignof__(id);
        case _C_STRUCT_B: //	'{'
	    return __alignof__(void*);
        case _C_STRUCT_E: //	'}'
            [NSException raise:@"Primitive's Size" format:@"don't know the size of the end of a structure."];
            return __alignof__(id);
        default:
	    printf("unknown type: %s\n",type);
            return __alignof__(id);
    }       
    return -1;
}
#endif

